#ifndef _MY_QUEUE_H
#define _MY_QUEUE_H

#include <stdatomic.h>

#include <spec_lib.h>
#include <stdlib.h>
#include <cdsannotate.h>
#include <specannotation.h>
#include <model_memory.h>
#include "common.h" 

#define MAX_NODES			0xf

typedef unsigned long long pointer;
typedef atomic_ullong pointer_t;

#define MAKE_POINTER(ptr, count)	((((pointer)count) << 32) | ptr)
#define PTR_MASK 0xffffffffLL
#define COUNT_MASK (0xffffffffLL << 32)

static inline void set_count(pointer *p, unsigned int val) { *p = (*p & ~COUNT_MASK) | ((pointer)val << 32); }
static inline void set_ptr(pointer *p, unsigned int val) { *p = (*p & ~PTR_MASK) | val; }
static inline unsigned int get_count(pointer p) { return (p & COUNT_MASK) >> 32; }
static inline unsigned int get_ptr(pointer p) { return p & PTR_MASK; }

typedef struct node {
	unsigned int value;
	pointer_t next;
} node_t;

typedef struct {
	pointer_t head;
	pointer_t tail;
	node_t nodes[MAX_NODES + 1];
} queue_t;

void init_queue(queue_t *q, int num_threads);

/**
	@Begin
	@Options:
		LANG = C;
	@Global_define:
		@DeclareVar:
			IntegerList *__queue;
		@InitVar:
			__queue = createIntegerList();
		@Finalize:
			if (__queue)
				destroyIntegerList(__queue);
			return true;
	@Happens_before: Enqueue -> Dequeue
	@Commutativity: Enqueue <-> Dequeue: true
	@Commutativity: Dequeue <-> Dequeue: !_Method1.__RET__ || !_Method2.__RET__
	//@Commutativity: Enqueue <-> Enqueue: _Method1.q != _Method2.q
	//@Commutativity: Dequeue <-> Dequeue: _Method1.q != _Method2.q
	@End
*/



/**
	@Begin
	@Interface: Enqueue
	@Commit_point_set: EnqueueUpdateNext
	@ID: val
	@Action:
		# __ID__ is an internal macro that refers to the id of the current
		# interface call
		push_back(__queue, val);
		//model_print("Enqueue: val=%d\n", val);
	@End
*/
void enqueue(queue_t *q, unsigned int val);

/**
	@Begin
	@Interface: Dequeue
	@Commit_point_set: DequeueReadHead | DequeueUpdateHead | DequeueReadNextVerify
	@ID: __RET__ ? *retVal : 0
	@Action:
		int elem = 0;
		if (__RET__) {
			elem = front(__queue);
			pop_front(__queue);
		}
		//model_print("Dequeue: __RET__=%d, retVal=%d, elem=%d, \n", __RET__, *retVal, elem);
		//model_print("Result: %d\n", __RET__ ? *retVal == elem : true);
	@Post_check: __RET__ ? *retVal == elem : true
	@End
*/
bool dequeue(queue_t *q, int *retVal);
int get_thread_num();

#endif
